################################################################################
# File: /pre_processor.py                                                      #
# Project: d01_pre_processing                                                  #
# File Created: Sunday, 10th October 2021 4:23:01 am                           #
# Author: Ubuntu# zjoukhadar@unimelb.edu.au                                    #
# ---------------------------------------------------------------              #
# Last Modified: Sunday, 10th October 2021 4:23:01 am                          #
# Modified By: Ubuntu# zjoukhadar@unimelb.edu.au                               #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

from ..d_utils import logger
from ..d_utils.runtime_context import RuntimeContext
from ..d_utils.enums import Step, Development_state
from .services import (
    initialise_srv,
    finalizing_srv,
    null_strategy_srv,
    prepare_data_for_modelling,
)
from ..d_utils.enums import Step
from ..d00_data_access.services import data_savers, data_getters


def run(rc: RuntimeContext):
    try:
        rc = initialise_srv.initialise_pre_processor(rc)
    except Exception as e:
        logger.exception("initializing pre_processing step failed. %s" % e.__str__())
        raise e
    if rc.config.is_pipeline_step_enabled(Step.pre_processed):
        logger.info("pre_processing is enabled")
        logger.info("pre_processing started running...")
        try:
            rc = null_strategy_srv.run(rc)
            if (
                rc.pre_processing_context.development_state
                != Development_state.skeleton
            ):
                rc = prepare_data_for_modelling.run(rc)
            rc = finalizing_srv.run(rc)
        except Exception as e:
            logger.exception("finalising pre_processing step failed. %s" % e.__str__())
            raise e
        logger.info("pre_processing finished running...")
        return rc
    else:
        logger.info("pre_processing step is disabled")
        return rc


def move_data_to_general(rc: RuntimeContext):

    return



################################################################################
# File: /deepeye/d01_pre_processing/services/initialise_srv.py                 #
# Project: deep_eye_no_dementia                                                #
# File Created: Wednesday, 11th May 2022 1:27:52 am                            #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 11th May 2022 1:27:52 am                           #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils.runtime_context import RuntimeContext

from ...d_utils.enums import *
from ...d_utils import logger
from pathlib import Path
import os


def initialise_pre_processor(rc: RuntimeContext):
    from ..model.contextmodel import PreProcessingContext

    logger.info("initializing pre_processing")

    my_pre_processor_context = PreProcessingContext()
    data_access_data_sets = rc.config.get_data_sets()
    pre_processing_data_sets = rc.config.get_data_sets(step=Step.pre_processed)

    my_pre_processor_context.data_sets = [
        da_ds
        for pp_ds in pre_processing_data_sets
        for da_ds in data_access_data_sets
        if da_ds.name == pp_ds
    ]
    my_pre_processor_context.data_browser = rc.data_browser
    my_pre_processor_context.mode = rc.config.get_mode(Step.pre_processed)
    my_pre_processor_context.development_state = rc.config.get_development_state()
    my_pre_processor_context.null_strategy = rc.config.get_null_strategy()
    try:
        my_pre_processor_context.check_it()
        my_pre_processor_context.make_it()
    except Exception as e:
        raise e

    rc.pre_processing_context = my_pre_processor_context

    Path(rc.pre_processing_context.get_datasets_path()).mkdir(
        parents=True, exist_ok=True
    )

    os.environ["TFDS_DATA_DIR"] = rc.pre_processing_context.get_datasets_path()
    return rc



################################################################################
# File: /enums.py                                                              #
# Project: d_utils                                                             #
# File Created: Sunday, 10th October 2021 4:15:22 am                           #
# Author: Ubuntu# zjoukhadar@unimelb.edu.au                                    #
# ---------------------------------------------------------------              #
# Last Modified: Sunday, 10th October 2021 4:15:22 am                          #
# Modified By: Ubuntu# zjoukhadar@unimelb.edu.au                               #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from enum import Enum


class Config_type(Enum):
    local = 1
    base = 2


class Step(Enum):
    data_access = 0
    pre_processed = 1
    modelled = 2
    outputted = 3


class Method(Enum):
    isomap = 1
    method01 = 2
    VisionTransformer = 3
    Conv2D = 4
    Conv3D = 5


class Problem_type(Enum):
    classification = 0
    segmentation = 0
    regression = 0


class Mode(Enum):
    development = 0
    production = 1


class Development_state(Enum):
    skeleton = 0
    in_progress = 2
    released = 1


class DataSetFormat(Enum):
    csv = 0
    images = 1



################################################################################
# File: /best_package/d02_modelling/model/my_network.py                        #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 7:45:33 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 7:45:33 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

from ...d_utils.runtime_context import RuntimeContext
from .hyper_parameters import HyperParameters
import tensorflow as tf
import functools
from .my_model import MyModel
from flax.metrics import tensorboard
from tqdm.auto import tqdm
import jax
import jax.numpy as jnp
import optax
from flax.training import train_state, checkpoints
from flax import jax_utils
from flax.training.common_utils import get_metrics


@functools.partial(
    jax.pmap, axis_name="batch", static_broadcasted_argnums=(4, 5)  # , backend="gpu"
)
def train_step(
    replicated_state,
    replicated_metrics,
    replicated_rng,
    batch,
    num_classes,
    train=True,
):
    def loss_fn(params, rng_key, images, labels):
        labels_onehot = jax.nn.one_hot(labels, num_classes=num_classes)
        rng, dropout_apply_rng = jax.random.split(rng_key)
        logits = replicated_state.apply_fn(
            {"params": params},
            images,
            train=train,
            rngs={"dropout": dropout_apply_rng},
        )
        loss = optax.softmax_cross_entropy(logits, labels_onehot).mean()
        acc = (logits.argmax(axis=-1) == labels).mean()
        return loss, acc

    imgs = batch["image"]
    labels = batch["label"]
    grad_fn = jax.value_and_grad(loss_fn, has_aux=True)

    # Get loss, gradients for loss, and other outputs of loss function
    (loss, acc), grads = grad_fn(replicated_state.params, replicated_rng, imgs, labels)

    gathered_loss = jax.lax.pmean(loss, axis_name="batch")
    gathered_acc = jax.lax.pmean(acc, axis_name="batch")
    # accuracy = jnp.mean(jnp.argmax(probs, -1) == labels)
    grads = jax.lax.pmean(grads, axis_name="batch")
    # Update parameters and batch statistics
    replicated_state = replicated_state.apply_gradients(grads=grads)
    replicated_metrics["loss"] = gathered_loss
    replicated_metrics["acc"] = gathered_acc
    return replicated_state, replicated_metrics


train_step = jax_utils.pad_shard_unpad(
    train_step, static_argnums=(0, 1, 2, 4, 5), static_return=True
)


def train_epoch(
    replicated_state,
    replicated_metrics,
    train_dataset,
    replicated_rng,
    num_classes,
    per_device_batch_size,
    train=True,
):
    nb_samples = (
        train_dataset.cardinality().numpy()
        * per_device_batch_size
        * jax.local_device_count()
    )

    samples_progressbar = tqdm(
        total=nb_samples,
        desc="Training... %i samples" % nb_samples,
        position=1,
        leave=False,
    )
    batches_metrics = []
    # Train model for one epoch, and log avg loss and accuracy
    for batch in train_dataset.as_numpy_iterator():
        replicated_state, updated_replicated_metrics = train_step(
            replicated_state,
            replicated_metrics,
            replicated_rng,
            batch,
            num_classes,
            train,
            min_device_batch=per_device_batch_size,
        )
        batches_metrics.append(updated_replicated_metrics)
        samples_progressbar.update(per_device_batch_size * jax.local_device_count())
    batches_metrics = get_metrics(batches_metrics)
    batches_metrics = jax.tree_map(jnp.mean, batches_metrics)
    samples_progressbar.close()
    train_loss = round(batches_metrics["loss"].item(), 4)
    train_accuracy = round(batches_metrics["acc"].item(), 4)
    return replicated_state, train_loss, train_accuracy


@functools.partial(
    jax.pmap, axis_name="batch", static_broadcasted_argnums=(4, 5)  # , backend="gpu"
)
def eval_step(
    replicated_state,
    replicated_metrics,
    replicated_rng,
    batch,
    num_classes,
    train=False,
):
    def loss_fn(params, rng_key, images, labels):
        labels_onehot = jax.nn.one_hot(labels, num_classes=num_classes)
        rng, dropout_apply_rng = jax.random.split(rng_key)
        logits = replicated_state.apply_fn(
            {"params": params},
            images,
            train=train,
            rngs={"dropout": dropout_apply_rng},
        )
        loss = optax.softmax_cross_entropy(logits, labels_onehot).mean()

        acc = (logits.argmax(axis=-1) == labels).mean()
        return loss, acc

    imgs = batch["image"]
    labels = batch["label"]
    loss, acc = loss_fn(replicated_state.params, replicated_rng, imgs, labels)

    gathered_loss = jax.lax.pmean(loss, axis_name="batch")
    gathered_acc = jax.lax.pmean(acc, axis_name="batch")

    replicated_metrics["loss"] = gathered_loss
    replicated_metrics["acc"] = gathered_acc
    return replicated_metrics


eval_step = jax_utils.pad_shard_unpad(
    eval_step, static_argnums=(0, 1, 2, 4, 5), static_return=True
)


def eval_model(
    replicated_state,
    replicated_metrics,
    test_dataset,
    replicated_rng,
    num_classes,
    per_device_batch_size,
    train=False,
):
    nb_samples = (
        test_dataset.cardinality().numpy()
        * per_device_batch_size
        * jax.local_device_count()
    )

    samples_progressbar = tqdm(
        total=nb_samples,
        desc="Testing... %i samples" % nb_samples,
        position=1,
        leave=False,
    )
    batches_metrics = []
    # Train model for one epoch, and log avg loss and accuracy
    for batch in test_dataset.as_numpy_iterator():
        updated_replicated_metrics = eval_step(
            replicated_state,
            replicated_metrics,
            replicated_rng,
            batch,
            num_classes,
            train,
            min_device_batch=per_device_batch_size,
        )
        batches_metrics.append(updated_replicated_metrics)
        samples_progressbar.update(per_device_batch_size * jax.local_device_count())

    batches_metrics = get_metrics(batches_metrics)
    batches_metrics = jax.tree_map(jnp.mean, batches_metrics)
    samples_progressbar.close()
    test_loss = round(batches_metrics["loss"].item(), 4)
    test_accuracy = round(batches_metrics["acc"].item(), 4)

    return test_loss, test_accuracy


def create_train_state(hyper_parameters: HyperParameters, input_shape, rng_key):

    init_rng, dropout_init_rng = jax.random.split(rng_key, 2)

    mymodel = MyModel(
        embed_dim=hyper_parameters.embed_dim,
        patch_size=hyper_parameters.patch_size,
        dropout_prob=hyper_parameters.dropout_prob,
    )
    params = mymodel.init(
        {"params": init_rng, "dropout": dropout_init_rng},
        jnp.ones(list(input_shape)),
        train=True,
    )["params"]

    optimizer = optax.chain(
        optax.clip_by_global_norm(1.0),  # Clip gradients at norm 1
        optax.adamw(
            learning_rate=hyper_parameters.learning_rate,
            weight_decay=hyper_parameters.weight_decay,
        ),
    )
    # Initialize training state
    state = train_state.TrainState.create(
        apply_fn=mymodel.apply,
        params=params,
        tx=optimizer,
    )
    return state


def train_and_evaluate(
    hyper_parameters: HyperParameters,
    runtime_context: RuntimeContext,
    train_dataset: tf.data.Dataset,
    val_dataset: tf.data.Dataset,
    seed=42,
):
    log_dir = runtime_context.modelling_context.get_modelling_folder_path()
    if jax.process_index() == 0:
        summary_writer = tensorboard.SummaryWriter(log_dir)

    rng_key = jax.random.PRNGKey(seed)

    input_shape = next(iter(train_dataset))["image"].shape

    state = create_train_state(
        hyper_parameters=hyper_parameters, input_shape=input_shape, rng_key=rng_key
    )

    replicated_state = jax_utils.replicate(state)

    replicated_metrics = jax_utils.replicate(
        dict(loss=jnp.array(0, jnp.int32), acc=jnp.array(0, jnp.int32))
    )

    val_replicated_metrics = jax_utils.replicate(
        dict(loss=jnp.array(0, jnp.int32), acc=jnp.array(0, jnp.int32))
    )
    replicated_rng = jax.random.split(rng_key, jax.local_device_count())

    epochs = tqdm(
        range(1, hyper_parameters.number_epochs), desc=f"Epoch ...", position=0
    )

    for epoch in epochs:

        replicated_state, train_loss, train_accuracy = train_epoch(
            replicated_state,
            replicated_metrics,
            train_dataset,
            replicated_rng,
            hyper_parameters.num_classes,
            runtime_context.modelling_context.per_device_batch_size,
            True,
        )

        epochs.write(
            "Epoch... (%i/%i | Train Loss: %f, Train Accuracy:%f"
            % (
                epoch,
                hyper_parameters.number_epochs,
                train_loss,
                train_accuracy,
            )
        )
        if epoch % 2 == 0:
            val_loss, val_accuracy = eval_model(
                replicated_state,
                val_replicated_metrics,
                val_dataset,
                replicated_rng,
                hyper_parameters.num_classes,
                runtime_context.modelling_context.per_device_batch_size,
                False,
            )
            epochs.write(
                "Epoch... (%i/%i | Test Loss: %f, Test Accuracy:%f"
                % (
                    epoch,
                    hyper_parameters.number_epochs,
                    val_loss,
                    val_accuracy,
                )
            )

        if jax.process_index() == 0:
            summary_writer.scalar("train/loss", train_loss, epoch)
            summary_writer.scalar("train/accuracy", train_accuracy, epoch)
            if epoch % 2 == 0:
                summary_writer.scalar("val/loss", val_loss, epoch)
                summary_writer.scalar("val/accuracy", val_accuracy, epoch)

    summary_writer.flush()
    return state



################################################################################
# File: /reportor.py                                                           #
# Project: d06_reporting                                                       #
# File Created: Sunday, 10th October 2021 4:24:08 am                           #
# Author: Ubuntu# zjoukhadar@unimelb.edu.au                                    #
# ---------------------------------------------------------------              #
# Last Modified: Sunday, 10th October 2021 4:24:08 am                          #
# Modified By: Ubuntu# zjoukhadar@unimelb.edu.au                               #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ..d_utils import logger

from ..d_utils.runtime_context import RuntimeContext
from ..d_utils.enums import Step
from .services import initialise_srv, finalizing_srv


def run(rc: RuntimeContext):
    try:
        rc = initialise_srv.initialise_outputting(rc)
    except Exception as e:
        logger.exception("initializing outputting step failed. %s" % e.__str__())
        raise e
    if rc.config.is_pipeline_step_enabled(Step.outputted):
        logger.info("outputting is enabled")

        try:
            rc = finalizing_srv.run(rc)
        except Exception as e:
            logger.exception("finalising outputting step failed. %s" % e.__str__())
            raise e
        logger.info("outputting finished running...")
    else:
        logger.info("outputting step is disabled")

    return rc



from .enums import *
from ..d00_data_access.model.datamodel import DataSet
import yaml


class ConfigController:
    runtime_base_path: str
    runtime_local_path: str

    def __init__(self, runtime_base_path, runtime_local_path):
        self.runtime_base_path = runtime_base_path
        self.runtime_local_path = runtime_local_path

    def __get_config(self, config_type: Config_type):
        if config_type == Config_type.base:
            config_path = self.runtime_base_path
        if config_type == Config_type.local:
            config_path = self.runtime_local_path
        with open(config_path, "r") as f:
            conf = yaml.safe_load(f)
            return conf

    def is_pipeline_step_enabled(self, ps: Step):
        return self.__get_config(Config_type.base)["pipeline"][ps.name]["enabled"]

    def should_save_data(self, ps: Step):
        return self.__get_config(Config_type.base)["pipeline"][ps.name]["save_data"]

    def get_data_sets(self, step: Step = None):
        if step:
            return self.__get_config(Config_type.base)["pipeline"][step.name][
                "data_sets"
            ].split(",")

        else:
            return [
                DataSet(
                    name=item["data_set"]["name"],
                    is_in_separate_folder=item["data_set"]["is_in_separate_folder"],
                )
                for item in self.__get_config(Config_type.base)["pipeline"][
                    Step.data_access.name
                ]
            ]

    def get_mode(self, step: Step):
        return Mode[self.__get_config(Config_type.base)["pipeline"][step.name]["mode"]]

    def get_development_state(self):
        return Development_state[
            self.__get_config(Config_type.base)["pipeline"]["development_state"]
        ]

    def get_method(self, step: Step):
        return Method[
            self.__get_config(Config_type.base)["pipeline"][step.name]["method"]
        ]

    def should_clear_last_run(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "clear_last_run_data"
        ]

    def keep_this_run_data(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "keep_data_of_run_name"
        ]

    def should_clear_last_output(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.outputted.name][
            "clear_last_output_data"
        ]

    def should_split(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "should_split"
        ]

    def what_run_name_to_use_in_output(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.outputted.name][
            "use_modelling_run_name"
        ]

    def keep_this_output_data(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "keep_output_data_of"
        ]

    def get_problem_type(self):
        return Problem_type[
            self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
                "problem_type"
            ]
        ]

    def get_number_of_classes(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "number_classes"
        ]

    def get_number_of_channels(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "number_channels"
        ]

    def get_per_device_batch_size(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "per_device_batch_size"
        ]

    def get_patch_size(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "patch_size"
        ]

    def get_test_size(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "test_size"
        ]

    def get_number_epochs(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.modelled.name][
            "epochs"
        ]

    def get_null_strategy(self):
        return self.__get_config(Config_type.base)["pipeline"][Step.pre_processed.name][
            "null_strategy"
        ]

    def get_input_variables(self):
        return self.__get_variables(input=True)

    def get_target_variables(self):
        return self.__get_variables(input=False)

    #############################
    ########SETTINGS#############
    #############################

    def get_sync_settings_folder_names(self):
        return self.__get_config(Config_type.base)["settings"]["data_sync"][
            "folder_names"
        ].split(",")

    def get_sync_settings_from(self):
        return self.__get_config(Config_type.base)["settings"]["data_sync"]["from"]

    def get_sync_settings_to(self):
        return self.__get_config(Config_type.base)["settings"]["data_sync"]["to"]

    def get_spartan_user_name(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["user_name"]

    def get_spartan_conda_env(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["conda_env"]

    def get_spartan_qos_setting(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["qos"]

    def get_spartan_account(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["account"]

    def get_spartan_partition(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["partition"]

    def get_spartan_cpus_per_task(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"][
            "cpus_per_task"
        ]

    def get_spartan_gpus_per_node(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"][
            "gpus_per_node"
        ]

    def get_spartan_memory(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["memory"]

    def get_spartan_time_days(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["time"][
            "days"
        ]

    def get_spartan_time_hours(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["time"][
            "hours"
        ]

    def get_spartan_time_minutes(self):
        return self.__get_config(Config_type.base)["settings"]["spartan"]["time"][
            "minutes"
        ]

    def get_spartan_data_dir(self):
        return self.__get_config(Config_type.local)["spartan_data_dir"]

    def get_mrc_data_dir(self):
        return self.__get_config(Config_type.local)["mrc_data_dir"]



################################################################################
# File: /best_package/d00_data_access/services/initialise_srv.py               #
# Project: best_project                                                        #
# File Created: Friday, 9th September 2022 2:06:00 am                          #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Friday, 9th September 2022 2:06:00 am                         #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

from ...d_utils.runtime_context import RuntimeContext
from ...d_utils import logger


def initialise_data_access(rc: RuntimeContext):
    logger.info("initializing data_access")
    from ..model.contextmodel import DataAccessContext

    my_data_access_context = DataAccessContext()
    my_data_access_context.data_sets = rc.config.get_data_sets()
    my_data_access_context.data_browser = rc.data_browser
    my_data_access_context.development_state = rc.config.get_development_state()
    try:
        my_data_access_context.run()
        my_data_access_context.make_it()
    except Exception as e:
        raise e

    rc.data_access_context = my_data_access_context

    return rc



################################################################################
# File: /configration.py                                                       #
# Project: d_utils                                                             #
# File Created: Monday, 12th April 2021 4:53:20 am                             #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Monday, 12th April 2021 4:53:21 am                            #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

import logging
from datetime import datetime

from .enums import Config_type, Method,  Step
from . import logger
import pathlib as pl
import yaml

import getpass
import shutil
from .. import get_version
from .. import get_package_name
import glob
import os
from .config_controller import ConfigController
from .data_browser import DataBrowser
from ..d00_data_access.model.contextmodel import DataAccessContext
from ..d01_pre_processing.model.contextmodel import PreProcessingContext
from ..d02_modelling.model.contextmodel import ModellingContext
from ..d03_outputting.model.contextmodel import OutputtingContext




class RuntimeContext:

    runtime_tag: str
    root_data_path: str
    original_data_path: str
    execution_path: str
    log_path: str
    is_notebook: bool
    package_name: str
    config: ConfigController
    data_browser: DataBrowser
    data_access_context : DataAccessContext
    pre_processing_context: PreProcessingContext
    modelling_context : ModellingContext
    outputting_context : OutputtingContext

    is_running_on_spartan: bool
    def __init__(self, execuation_path: str, is_notebook: bool):
        self.runtime_tag = (
            get_version() + "_" + datetime.now().strftime("%d-%m-%Y_%H_%M_%S")
        )
        self.execution_path = execuation_path
        self.is_notebook = is_notebook
        self.package_name = get_package_name()
        self.is_running_on_spartan = False
        self.store_execution_path()

        self.copy_runtime_base_config_file()

        self.copy_runtime_local_config_file()

        local_config_path = self.get_runtime_local_config_file_path()

        with open(local_config_path, "r") as f:
            local_conf = yaml.safe_load(f)
            self.original_data_path = local_conf["original_data_path"]
            self.root_data_path = local_conf["root_data_path"]

        self.config = ConfigController(self.get_runtime_base_config_file_path(), self.get_runtime_local_config_file_path())

  
        try:
            self.data_browser = DataBrowser(
                original_data_path= self.original_data_path, root_data_path= self.root_data_path
                )
        except Exception as e:
            logger.exception(e)
    
    def run_configuration(self):

        self.setup_logger()
    
        return

    def setup_logger(self):
        import jax
        logger = logging.getLogger(self.package_name)
        #logger.setLevel(logging.DEBUG)
        # We only want one process per machine to log things on the screen
        logger.setLevel(logging.INFO if jax.process_index() == 0 else logging.ERROR)
        formatter = logging.Formatter(
            "[%(asctime)s|%(levelname)s|%(processName)s|%(threadName)s| %(message)s"
        )

        handler = logging.FileHandler(self.get_log_file_path())
        stout_handler = logging._StderrHandler(logging.DEBUG)
        stout_handler.setFormatter(formatter)
        handler.setFormatter(formatter)

        # this bit will make sure we won't have
        # duplicated messages in the output
        if not len(logger.handlers):
            logger.addHandler(handler)
            logger.addHandler(stout_handler)

    def store_execution_path(self):
        with open(self.get_runtime_data_file_path(), "w") as file:
            documents = yaml.dump({"package_execution_path": str(self.execution_path)}, file)

    def get_runtime_data_file_path(self):
        return self.get_runtime_package_data_dir_path() + "/runtime_data.yml"

    def get_runtime_package_data_dir_path(self):
        curr_user_home_dir = str(pl.Path.home())
        package_data_dir_path = (
            curr_user_home_dir + "/."+self.package_name+"_package_data/" + self.runtime_tag
        )
        pl.Path(package_data_dir_path).mkdir(parents=True, exist_ok=True)
        return package_data_dir_path

    def get_log_file_path(self):
        log_dir_path = str(pl.Path.home()) + "/."+self.package_name+"_package_data/" + "logs/"
        pl.Path(log_dir_path).mkdir(parents=True, exist_ok=True)
        return log_dir_path + self.package_name+".%s" % self.runtime_tag + ".log"


   
    
    
    def copy_runtime_base_config_file(self):
        try:
            with open(
                self.get_package_execution_path() + "/conf/base.yaml"
            ) as yamlfile:
                base_config = yaml.safe_load(yamlfile)
                if base_config:
                    with open(
                        self.get_runtime_package_data_dir_path() + "/base.yaml", "w"
                    ) as yamlfile:
                        yaml.safe_dump(base_config, yamlfile)
        except Exception as e:
            logger.exception(
                "No base.yaml was found. Please create base.yaml at /conf/base.yaml. See documentation for more info."
            )
            raise e

    def copy_runtime_local_config_file(self):
        try:
            with open(
                self.get_package_execution_path() + "/conf/local.yaml"
            ) as yamlfile:
                local_config = yaml.safe_load(yamlfile)
                if local_config:
                    with open(
                        self.get_runtime_package_data_dir_path() + "/local.yaml", "w"
                    ) as yamlfile:
                        yaml.safe_dump(local_config, yamlfile)
        except Exception as e:
            logger.exception(
                "No local.yaml was found. Please create local.yaml at /conf/local.yaml. See documentation for more info."
            )
            raise e

    def get_runtime_base_config_file_path(self):
        return self.get_runtime_package_data_dir_path() + "/base.yaml"

    def get_runtime_local_config_file_path(self):
        return self.get_runtime_package_data_dir_path() + "/local.yaml"

    def get_package_execution_path(self):
        try:
            with open((self.get_runtime_data_file_path()), "r") as stream:
                try:
                    app_data = yaml.safe_load(stream)
                    return app_data["package_execution_path"]
                except yaml.YAMLError as exc:
                    logger.info(
                        "Cannot read runtime_data.yml file. Message: " + exc
                    )
        except FileNotFoundError as fnf_error:
            logger.exception(fnf_error)

    def write_to_runtime_data(self, key, value):
        try:
            with open((self.get_runtime_data_file_path()), "r") as stream:
                app_data = yaml.safe_load(stream)
                app_data[key] = value
            with open((self.get_runtime_data_file_path()), "w") as stream:
                yaml.safe_dump(app_data, stream)
        except FileNotFoundError as fnf_error:
            logger.exception(fnf_error)

    def read_from_runtime_data(self, key):
        try:
            with open((self.get_runtime_data_file_path()), "r") as stream:
                app_data = yaml.safe_load(stream)
                return app_data[key]
        except FileNotFoundError as fnf_error:
            logger.exception(fnf_error)

    def clean(self):
        # remove temporary app data when this run is finished
        try:
            shutil.rmtree(self.get_runtime_package_data_dir_path())
            # remove test logs
            if hasattr(self, "root_data_path"):
                if self.root_data_path.endswith("test/"):
                    shutil.rmtree(self.root_data_path)
            log_dir_path = str(pl.Path.home()) + "/."+self.package_name+"_package_data/" + "logs/"
            for f in glob.glob(log_dir_path + "*_test.log"):
                os.remove(f)
        except Exception as e:
            logger.exception(e)

    

    def get_previous_run_name(self, step: Step):
        try:
            #pick the lasted package app run time data
            curr_user_home_dir = str(pl.Path.home())
            package_data_dir_path = curr_user_home_dir + "/."+self.package_name+"_package_data/"
            sortedList = sorted(glob.glob(os.path.join(package_data_dir_path, '[!logs]*/')), key=os.path.getmtime)
            if len(sortedList) > 1:
                for i in range(len(sortedList) +1 ):
                    if i < 2:
                        continue
                    else:
                        pre_path = sortedList[-i] + "runtime_data.yml"
                        try:
                            with open(pre_path, "r") as stream:
                                app_data = yaml.safe_load(stream)
                                if step.name+"_run_name" in app_data:
                                    return app_data[step.name+"_run_name"]
                        except FileNotFoundError as fnf_error:
                            continue
                return 0
            else:
                return 0
        except Exception as e:  
            logger.exception(e)
            return 0

    def get_current_run_name(self, step: Step):
        try:
            #pick the lasted package app run time data
            with open((self.get_runtime_data_file_path()), "r") as stream:
                app_data = yaml.safe_load(stream)
                if not step.name+"_run_name" in app_data:
                    return 0
                else:
                    return app_data[step.name+"_run_name"]
        except FileNotFoundError as fnf_error:
            logger.exception(fnf_error)
            return 0

    def set_current_run_name(self, run_name: str, step: Step):
        try:
            with open((self.get_runtime_data_file_path()), "r") as stream:
                app_data = yaml.safe_load(stream)
                app_data[step.name+"_run_name"] = run_name
            with open((self.get_runtime_data_file_path()), "w") as stream:
                yaml.safe_dump(app_data, stream)
        except FileNotFoundError as fnf_error:
            logger.exception(fnf_error)
    
   


################################################################################
# File: /best_package/d01_pre_processing/services/prepare_data_for_modelling.py#
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 7:19:24 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 7:19:24 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils.runtime_context import RuntimeContext

from ...d00_data_access.datasets import dataset01_dataset


def run(rc: RuntimeContext):
    import tensorflow_datasets as tfds

    data_path = rc.pre_processing_context.get_working_data_path()
    dl_config = tfds.download.DownloadConfig(manual_dir=data_path + "metadata")
    db = tfds.builder("dataset01_dataset", rc=rc)
    db.download_and_prepare(download_config=dl_config)
    return rc



import logging
import sys
from pathlib import Path
import os


def get_package_name():
    return os.path.basename(str(Path(__file__).parents[1]))


def info(message: str):
    return logging.getLogger(get_package_name()).info(message)


def warning(message: str):
    return logging.getLogger(get_package_name()).warning(message)


def debug(message: str):
    return logging.getLogger(get_package_name()).debug(message)


def error(message: str):
    return logging.getLogger(get_package_name()).error(message)


def exception(message: str):
    return logging.getLogger(get_package_name()).exception(message)


def log(message: str):
    return logging.getLogger(get_package_name()).log(message)


def critical(message: str):
    return logging.getLogger(get_package_name()).critical(message)



from ...d_utils.enums import Development_state, Method, Mode, Problem_type
from ...d00_data_access.model.datamodel import *
from typing import List
from ...d_utils.data_browser import DataBrowser
import os
import pathlib as pl


class ModellingContext:
    data_sets: List[DataSet]
    data_browser: DataBrowser
    current_run_name: str
    last_run_name: str
    mode: Mode
    method: Method
    development_state: Development_state
    test_size: float
    problem_type: Problem_type
    number_classes: int
    number_epochs: int
    per_device_batch_size: int
    patch_size: int
    number_channels: int

    def __init__(self, current_run_name: str, last_run_name: str):
        self.current_run_name = current_run_name
        self.last_run_name = last_run_name
        pass

    def make_modelling_folder(self):
        try:
            pl.Path(
                self.data_browser.run_modelling_path(self.mode, self.current_run_name)
            ).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise e
        return 0

    def get_modelling_folder_path(self, run_name: str = None):
        if run_name:
            temp_run_name = run_name
        else:
            temp_run_name = self.current_run_name
        return self.data_browser.run_modelling_path(self.mode, temp_run_name)

    def check_it(self):
        if self.development_state != Development_state.skeleton:
            pre_processed_data_path = (
                self.data_browser.step_working_path(Step.pre_processed, self.mode)
                + "datasets"
            )
            if not any(os.scandir(pre_processed_data_path)):
                raise RuntimeError(
                    "cannot do modelling, no pre-processed data found. Please run data pre-processing step first"
                )

            return True

    def make_it(self):
        self.make_modelling_folder()

        return 0



from glob import glob
from .d00_data_access import data_access
from .d01_pre_processing import pre_processor
from .d02_modelling import modeller
from .d03_outputting import outputter
from .d_utils import spartan
from .d_utils.runtime_context import RuntimeContext
from .d_utils import logger
import os


def start(rc: RuntimeContext):

    rc.write_to_runtime_data("engine_started", True)
    logger.info("%s engine has started" % get_package_name())


def run_data_sync(rc: RuntimeContext):
    data_access.run_data_sync(rc)


def run_on_spartan(rc: RuntimeContext):
    spartan.generate_and_submit_spartan_job(rc)


def run(rc: RuntimeContext):

    if rc.read_from_runtime_data("engine_started"):
        logger.info("%s engine is now running." % get_package_name())
        try:
            rc = data_access.run(rc)
            rc = pre_processor.run(rc)
            rc = modeller.run(rc)
            rc = outputter.run(rc)
        except Exception as e:
            logger.error(e.__str__())
            raise e

    else:
        logger.error(
            "Engine has not been started. Please start the engine before running it by calling start()"
        )


def get_package_name():
    folder_path = os.path.dirname(__file__)
    package_name = os.path.basename(folder_path)
    return package_name


if __name__ == "__main__":
    run()



################################################################################
# File: /best_package/d02_modelling/model/my_model.py                          #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 8:22:23 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 8:22:23 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

from flax import linen as nn


class MyModel(nn.module):
    embd_dim: int
    patch_size: int
    dropout_prob: float
    num_classes: int

    def setup(self):
        self.input_layer = nn.Dense(self.embd_dim)
        self.dropout = nn.Dropout(self.dropout_prob)
        self.mlp_head = nn.Sequential([nn.LayerNorm(), nn.Dense(self.num_classes)])

    def __call__(self, x, train=True):
        inp_x = self.img_to_patch(x, self.patch_size)
        B, T, _ = x.shape

        out = self.input_layer(x)
        out = self.dropout(x, deterministic=not train)

        out = self.mlp_head(out)
        return out

    def img_to_patch(self, x, patch_size, flatten_channels=True):
        """
        Inputs:
            x - torch.Tensor representing the image of shape [B, H, W, C]
            patch_size - Number of pixels per dimension of the patches (integer)
            flatten_channels - If True, the patches will be returned in a flattened format
                            as a feature vector instead of a image grid.
        """
        B, H, W, C = x.shape
        x = x.reshape(B, H // patch_size, patch_size, W // patch_size, patch_size, C)
        x = x.transpose(0, 1, 3, 2, 4, 5)  # [B, H', W', p_H, p_W, C]
        x = x.reshape(B, -1, *x.shape[3:])  # [B, H'*W', p_H, p_W, C]
        if flatten_channels:
            x = x.reshape(B, x.shape[1], -1)  # [B, H'*W', p_H*p_W*C]
        return x



################################################################################
# File: /deepeye/d00_data_arc.configess/services/spartan_mrc_data_sync_srv.py         #
# Project: deep_eye_no_dementia                                                #
# File Created: Wednesday, 11th May 2022 1:53:29 am                            #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 11th May 2022 1:53:29 am                           #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils import logger
import subprocess
from ...d_utils.runtime_context import RuntimeContext
import getpass
from ... import get_version


def constrcut_command_list(rc: RuntimeContext):

    source = rc.config.get_sync_settings_from()
    dest = rc.config.get_sync_settings_to()
    list_folders = rc.config.get_sync_settings_folder_names()
    list_commands = []
    source_data_path = ""
    dest_data_path = ""
    spartan = rc.config.get_spartan_data_dir()
    mrc = rc.config.get_mrc_data_dir()
    if spartan == "":
        logger.error(
            "Spartan data location is not provided. Make sure the key 'spartan_data_dir' in local.yaml is poiting to where the data is stored on spartan"
        )
        return 1
    if mrc == "":
        logger.error(
            "mrc data location is not provided. Make sure the key 'mrc_data_dir' in local.yaml is poiting to where the data is stored on mrc vm"
        )
        return 1

    if source == "spartan":
        source_data_path = rc.config.get_spartan_data_dir()
        if dest == "mrc":
            dest_data_path = rc.config.get_mrc_data_dir()
        for fn in list_folders:
            # if fn == 'original': continue
            list_commands.append(
                "rsync -avz  "
                + rc.config.get_spartan_user_name()
                + "@spartan.hpc.unimelb.edu.au:"
                + source_data_path
                + "/"
                + fn
                + " "
                + dest_data_path
                + "/"
            )
        return list_commands
    if dest == "spartan":
        if source == "mrc":
            source_data_path = rc.config.get_mrc_data_dir()
        dest_data_path = rc.config.get_spartan_data_dir()
        for fn in list_folders:
            # if fn == 'original': continue
            # list_commands.append(
            # "rsync -avz  "
            # + rc.config.get_spartan_user_name()
            # + "@spartan.hpc.unimelb.edu.au:"
            # + source_data_path
            # + "/"
            # + fn
            # + " "
            # + dest_data_path
            # + "/"
            # )
            list_commands.append(
                "rsync -avz  "
                + source_data_path
                + getpass.getuser()
                + "/deepeye/"
                + get_version()
                + "/"
                + fn
                + "/oct/02_processed/datasets"
                + " "
                + rc.config.get_spartan_user_name()
                + "@spartan.hpc.unimelb.edu.au:"
                + dest_data_path
                + "derived/oct/02_processed/"
            )
        return list_commands


def run_command(command):
    completed = subprocess.run(command, shell=True)
    print("returncode:", completed.returncode)
    return


def execute_data_sync(rc: RuntimeContext):
    # rc.config.save_run_time_configs()
    source = rc.config.get_sync_settings_from()
    dest = rc.config.get_sync_settings_to()
    list_folders = rc.config.get_sync_settings_folder_names()
    if source == "spartan":
        source_path = rc.config.get_spartan_data_dir()
        dest_path = rc.config.get_mrc_data_dir()
    if source == "mrc":
        source_path = rc.config.get_mrc_data_dir()
        dest_path = rc.config.get_spartan_data_dir()

    logger.info(
        "Moving data from folders %s at %s in %s to %s at %s"
        % (" ".join(list_folders), source_path, source, dest, dest_path)
    )
    command_list = constrcut_command_list(rc)
    if command_list == 1:
        return
    for comm in command_list:
        run_command(comm)



from ..d_utils.runtime_context import RuntimeContext
from ..d_utils.enums import *
import shutil
import os
from pathlib import Path


def clear_last_run_data(rc: RuntimeContext, step: Step):
    if step not in [Step.modelled, Step.outputted]:
        raise RuntimeError("There is no run folder for %s" % step.name)

    if step == Step.modelled:
        last_run_dir = rc.data_browser.run_modelling_path(
            rc.modelling_context.mode, rc.modelling_context.last_run_name
        )
    else:
        last_run_dir = rc.data_browser.run_output_path(
            rc.modelling_context.mode, rc.modelling_context.last_run_name
        )

    if os.path.isdir(last_run_dir):
        shutil.rmtree(last_run_dir)
    # delete run folder if empty
    last_run_dir_path = Path(last_run_dir).parent
    is_empty = Path(last_run_dir_path).is_dir() and not any(
        Path(last_run_dir_path).iterdir()
    )
    if is_empty:
        shutil.rmtree(str(last_run_dir_path))

    return



################################################################################
# File: /best_package/d01_pre_processing/services/finishing_srv.py             #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 5:39:39 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 5:39:39 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils.runtime_context import RuntimeContext
from ...d_utils import logger
import pickle
from ...d_utils import file_names


def run(rc: RuntimeContext):
    logger.info("finalizing pre_processing step")
    where_to_save = rc.pre_processing_context.get_working_data_path()
    full_path = where_to_save + file_names.pre_processing_context_file_name()
    with open(full_path, "wb") as handle:
        pickle.dump(rc.pre_processing_context, handle, protocol=pickle.HIGHEST_PROTOCOL)
    return rc



################################################################################
# File: /data_access.py                                                        #
# Project: d00_data_access                                                     #
# File Created: Sunday, 10th October 2021 4:22:52 am                           #
# Author: Ubuntu# zjoukhadar@unimelb.edu.au                                    #
# ---------------------------------------------------------------              #
# Last Modified: Sunday, 10th October 2021 4:22:52 am                          #
# Modified By: Ubuntu# zjoukhadar@unimelb.edu.au                               #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ..d_utils import logger
from ..d_utils.runtime_context import RuntimeContext
from .services import spartan_mrc_data_sync_srv
from .services import initialise_srv


def run(rc: RuntimeContext):
    logger.info("running data_access")
    try:
        rc = initialise_srv.initialise_data_access(rc)
    except Exception as e:
        logger.exception("initializing data access has failed. %s" % e.__str__())
        raise e
    return rc


def run_data_sync(rc: RuntimeContext):
    spartan_mrc_data_sync_srv.execute_data_sync(rc)



################################################################################
# File: /best_package/d03_outputting/model/__pycache__/contextmodel.py         #
# Project: best_project                                                        #
# File Created: Friday, 9th September 2022 6:19:47 am                          #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Friday, 9th September 2022 6:19:47 am                         #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

from ...d_utils.enums import Development_state, Method, Mode
from ...d00_data_access.model.datamodel import *
from typing import List
from ...d_utils.data_browser import DataBrowser
import os
import pathlib as pl


class OutputtingContext:
    data_sets: List[DataSet]
    data_browser: DataBrowser
    modeller_run_name: str
    last_run_name: str
    mode: Mode
    method: Method
    development_state: Development_state

    def __init__(self, modeller_run_name: str, last_run_name):
        self.modeller_run_name = modeller_run_name
        self.last_run_name = last_run_name
        pass

    def check_modelled_data_exist(self):

        data_path = self.data_browser.run_modelling_path(
            self.mode, self.modeller_run_name
        )
        # check if there is data there
        data_size = 0
        for dirpath, dirnames, filenames in os.walk(data_path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if not os.path.islink(fp):
                    data_size += os.path.getsize(fp)
        if data_size == 0:
            raise RuntimeError(
                "modelled data for %s is empty. Make sure that you have the data for modelled %s placed at %s"
                % (self.modeller_run_name, self.modeller_run_name, data_path)
            )
        return True

    def make_outputting_folder(self):
        try:
            for data_set in self.data_sets:
                pl.Path(
                    self.data_browser.run_output_path(
                        self.mode, self.modeller_run_name, data_set
                    )
                ).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise e
        return 0

    def run(self):
        if self.development_state != Development_state.skeleton:
            self.check_modelled_data_exist()
        return True

    def make_it(self):
        self.make_outputting_folder()

        return 0



"""ukbiobank_dataset dataset."""

import tensorflow_datasets as tfds
import csv
from zipfile import ZipFile
import io
from ....d_utils.runtime_context import RuntimeContext
import numpy as np
from PIL import Image
import tensorflow as tf

# TODO(ukbiobank_dataset): Markdown description  that will appear on the catalog page.
_DESCRIPTION = """
Description is **formatted** as markdown.

It should also contain any processing which has been applied (if any),
(e.g. corrupted example skipped, images cropped,...):
"""

# TODO(ukbiobank_dataset): BibTeX citation
_CITATION = """
"""


class Dataset01Dataset(tfds.core.GeneratorBasedBuilder):
    """DatasetBuilder for ukbiobank_dataset dataset."""

    MANUAL_DOWNLOAD_INSTRUCTIONS = """ UKBIOBANK Manual Dataset
    """
    VERSION = tfds.core.Version("1.0.0")
    RELEASE_NOTES = {
        "1.0.0": "Initial release.",
    }
    rc: RuntimeContext = None

    def __init__(self, rc: RuntimeContext) -> None:
        self.rc = rc
        super().__init__()

    def _info(self) -> tfds.core.DatasetInfo:
        """Returns the dataset metadata."""
        # TODO(ukbiobank_dataset): Specifies the tfds.core.DatasetInfo object
        return tfds.core.DatasetInfo(
            builder=self,
            description=_DESCRIPTION,
            features=tfds.features.FeaturesDict(
                {
                    # These are the features of your dataset like images, labels ...
                    "image": tfds.features.Image(
                        shape=(
                            512,
                            512,
                            1,
                        ),
                        dtype=tf.float32,
                    ),
                    "label": tfds.features.ClassLabel(num_classes=10),
                }
            ),
            # If there's a common (input, target) tuple from the
            # features, specify them here. They'll be used if
            # `as_supervised=True` in `builder.as_dataset`.
            supervised_keys=("image", "label"),  # Set to `None` to disable
            homepage="https://dataset-homepage/",
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager: tfds.download.DownloadManager):

        """Returns SplitGenerators."""
        # TODO(ukbiobank_dataset): Downloads the data and defines the splits
        # path = dl_manager.download_and_extract("https://todo-data-url")
        path = dl_manager.manual_dir

        # TODO(ukbiobank_dataset): Returns the Dict[split names, Iterator[Key, Example]]
        return {
            "dataset01": self._generate_examples(path / "dataset01.csv"),
        }

    def _generate_examples(self, path):
        """Yields examples."""
        with path.open() as f:
            idx = 0
            for row in csv.DictReader(f):
                idx = idx + 1
                zip_file = ZipFile(row["eyeleft_t0"])

                image_slice_data = zip_file.read(zip_file.namelist()[65])

                p_id = row["eid"]
                image_data = Image.open(io.BytesIO(image_slice_data))
                image = np.array(image_data, dtype=np.float32)
                image = image.reshape(image.shape[0], image.shape[1], 1)
                label = row["target"]
                # TODO(ukbiobank_dataset): Yields (key, example) tuples from the dataset
                yield p_id, {
                    "image": image,
                    "label": label,
                }



################################################################################
# File: /best_package/d02_modelling/model/hyper_parameters.py                  #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 8:05:53 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 8:05:53 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from dataclasses import dataclass


@dataclass
class HyperParameters:
    embed_dim: int = 64
    hidden_dim: int = 512
    num_heads: int = 8
    num_layers: int = 6
    patch_size: int = 512
    num_channels: int = 1
    num_patches: int = 512
    num_classes: int = 2
    number_epochs: int = 200
    dropout_prob: float = 0.2
    learning_rate: float = 3e-4
    weight_decay: float = 0.01



################################################################################
# File: /deepeye/d_utils/spartan.py                                            #
# Project: deep_eye_no_dementia                                                #
# File Created: Wednesday, 11th May 2022 2:49:00 am                            #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 11th May 2022 2:49:01 am                           #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from datetime import datetime
from pathlib import Path

from . import tools as tl
from .runtime_context import RuntimeContext

import subprocess
import os


def constrcut_job_slurm_file(rc: RuntimeContext):
    datnow = datetime.now().strftime("%d-%m-%Y_%H:%M:%S")
    sparatn_dir = rc.execution_path
    # sparatn_dir = os.path.dirname(execution_path)
    Path(sparatn_dir + "/spartan_jobs").mkdir(parents=True, exist_ok=True)
    Path(sparatn_dir + "/spartan_outputs").mkdir(parents=True, exist_ok=True)
    filepath = sparatn_dir + "/spartan_jobs/job_%s.%s" % (datnow, "slurm")
    with open(filepath, "w") as f:
        f.write("#!/bin/bash\n")
        f.write(
            "# summary: %s, %s, \n"
            % (rc.config.get_data_set().name, rc.config.get_main_method().name)
        )
        f.write(
            "# epochs: %s, data sets included in training:[%s] ,\n"
            % (
                rc.config.get_number_epochs(),
                ", ".join(map(str, rc.config.get_data_set().name)),
            )
        )
        f.write(
            "# Created by the MDAP unimelb deepeye job script generator for SLURM\n"
        )
        f.write("# %s\n" % datnow)
        f.write("\n\n")
        f.write("# Partition for the job:\n")
        f.write("#SBATCH --partition=" + rc.config.get_spartan_partition() + "\n")
        f.write("\n\n")
        f.write("# Multithreaded (SMP) job: must run on one node\n")
        f.write("#SBATCH --nodes=1\n")
        f.write("#SBATCH -q " + rc.config.get_spartan_qos_setting() + "\n")
        f.write("\n\n")
        f.write("# The name of the job:\n")
        f.write('#SBATCH --job-name="deepeye_%s"\n' % datnow)
        f.write(
            "#SBATCH -o "
            + sparatn_dir
            + "/spartan_outputs/slurm_job_%s.%s" % (datnow, "out")
        )
        f.write("\n\n")
        f.write("# The project ID which this job should run under:\n")
        f.write('#SBATCH --account="' + rc.config.get_spartan_account() + '"\n')
        f.write("\n\n")
        f.write("# Maximum number of tasks/CPU cores used by the job:\n")
        f.write("#SBATCH --ntasks=1\n")
        f.write(
            "#SBATCH --cpus-per-task="
            + str(rc.config.get_spartan_cpus_per_task())
            + "\n"
        )
        f.write("\n\n")
        f.write("# Number of GPUs requested per node:\n")
        f.write(
            "#SBATCH --gres=gpu:" + str(rc.config.get_spartan_gpus_per_node()) + "\n"
        )
        f.write("\n\n")
        f.write("# The amount of memory in megabytes per process in the job:\n")
        f.write("#SBATCH --mem=" + str(rc.config.get_spartan_memory()) + "\n")
        f.write("\n\n")
        f.write("# The maximum running time of the job in days-hours:mins:sec\n")
        f.write(
            "#SBATCH --time="
            + str(rc.config.get_spartan_time_days())
            + "-"
            + str(rc.config.get_spartan_time_hours())
            + ":"
            + str(rc.config.get_spartan_time_minutes())
            + ":00\n"
        )
        f.write("\n\n")
        f.write("# check that the script is launched with sbatch\n")
        f.write('if [ "x$SLURM_JOB_ID" == "x" ]; then\n')
        f.write(
            'echo "You need to submit your job to the queuing system with sbatch"\n'
        )
        f.write("exit 1\n")
        f.write("fi\n")
        f.write("\n\n")
        f.write("# Run the job from the directory where it was launched (default)\n")
        f.write("\n\n")
        f.write("# The modules to load\n")

        f.write("module load cuda/11.7.0\n")
        f.write("module load cudnn/8.2.1.32-cuda-11.3.1\n")
        f.write("\n\n")
        f.write("# The job command(s):\n")
        f.write("#source activate " + rc.config.get_spartan_conda_env() + "\n")
        f.write("conda info --envs\n")

        # which python
        # python --version
        f.write("time python " + sparatn_dir + "/main.py spartan")
    return filepath


def run_job(job_file_name):
    completed = subprocess.run("sbatch " + job_file_name, shell=True)
    print("returncode:", completed.returncode)
    return


def generate_and_submit_spartan_job(rc: RuntimeContext):
    job_slurm_file_name = constrcut_job_slurm_file(rc)
    run_job(job_slurm_file_name)


if __name__ == "__main__":
    job_slurm_file_name = constrcut_job_slurm_file()
    run_job(job_slurm_file_name)



################################################################################
# File: __init__.py                                                            #
# Path: \__init__.py                                                           #
# File Created: Saturday, 4th January 2020 2:47:44 pm                          #
# Author: Zaher Joukhadar                                                      #
# zjoukhadar@unimelb.edu.au                                                    #
# ---------------------------------------------------------------              #
# Last Modified: Tuesday, 7th January 2020 5:21:01 pm                          #
# Modified By: Zaher Joukhadar                                                 #
# zjoukhadar@unimelb.edu.au                                                    #
# ---------------------------------------------------------------              #
# Copyright 2019 - 2020, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
import os

import sys
import pathlib as pl
from .d_utils import logger
from .d_utils.enums import *
from pathlib import Path
from pyfiglet import Figlet
from termcolor import colored


def initialise():
    f = Figlet(font="slant")
    print(colored(f.renderText(get_package_name()), color="green"))
    from .d_utils import package_initializer

    print("Initializing %s" % get_package_name())
    path = Path(os.path.abspath(sys.argv[0])).parents[0]

    try:
        package_initializer.run(path)
        print("%s has been initialized" % get_package_name())
    except Exception as e:
        print(e.__str__())
    return


def start(is_on_spartan: bool = False):
    from .d_utils import tools as tl
    from . import best_package_engine

    from .d_utils.runtime_context import RuntimeContext

    if not is_on_spartan:
        os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count=32"
    else:
        import tensorflow as tf

        tf.config.experimental.set_visible_devices([], "GPU")

    path = ""
    is_running_in_notebook = tl.isnotebook()
    if is_running_in_notebook:

        path = Path(os.getcwd()).parents[0]

        # tl.copy_config_file_when_run_in_notebook(Config_type.base)
        # tl.copy_config_file_when_run_in_notebook(Config_type.local)
    else:
        path = Path(os.path.abspath(sys.argv[0])).parents[0]

    try:
        runtime_context = RuntimeContext(
            execuation_path=path, is_notebook=is_running_in_notebook
        )
        runtime_context.is_running_on_spartan = is_on_spartan
        runtime_context.run_configuration()
        best_package_engine.start(runtime_context)

        return runtime_context
    except Exception as e:
        logger.error(e.__str__())


def run(runtime_context):
    from . import best_package_engine

    logger.info("running %s engine" % get_package_name())
    best_package_engine.run(runtime_context)


def run_data_sync(runtime_context):
    from . import best_package_engine

    logger.info("running data sync")
    best_package_engine.run_data_sync(runtime_context)


def run_on_spartan(runtime_context):
    from . import best_package_engine

    logger.info("generating and submitting job to spartan")
    best_package_engine.run_on_spartan(runtime_context)


def get_version():
    return __version__


def get_package_name():
    folder_path = os.path.dirname(__file__)
    package_name = os.path.basename(folder_path)
    return package_name


__version__ = "0.0.1.dev"



################################################################################
# File: /deepeye/d01_pre_processing/model/contextmodel.py                      #
# Project: deep_eye_no_dementia                                                #
# File Created: Thursday, 14th July 2022 4:35:49 am                            #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Thursday, 14th July 2022 4:35:50 am                           #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d00_data_access.model.datamodel import *
from typing import List
from ...d_utils.data_browser import DataBrowser
import os
import pathlib as pl
from ...d_utils.enums import *


class PreProcessingContext:

    data_sets: List[DataSet]
    data_browser: DataBrowser
    mode: Mode
    development_state: Development_state
    null_strategy: str

    def __init__(self):

        pass

    def make_preprocessing_folder(self):
        try:
            for data_set in self.data_sets:
                pl.Path(
                    self.data_browser.step_working_path(
                        Step.pre_processed, self.mode, data_set
                    )
                ).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            raise e
        return 0

    def get_working_data_path(self, data_set: DataSet = None):
        return self.data_browser.step_working_path(
            Step.pre_processed, self.mode, data_set
        )

    def get_datasets_path(self):
        return self.get_working_data_path() + "datasets"

    def check_it(self):

        # self.check_input_datasets_exist()

        return True

    def make_it(self):
        self.make_preprocessing_folder()
        return 0



################################################################################
# File: /best_package/d02_modelling/services/deeplearning_srv.py               #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 7:44:39 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 7:44:39 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils.runtime_context import RuntimeContext


import numpy as np


from ..model.my_trainer import train_and_evaluate
from ..model.hyper_parameters import HyperParameters


def run(rc: RuntimeContext):
    import tensorflow_datasets as tfds
    import jax

    db = tfds.builder("dataset01_dataset", rc=rc)
    dataset = db.as_dataset(split=tfds.split_for_jax_process("dataset01"))
    train_size = int(1 - rc.modelling_context.test_size * len(dataset))
    val_size = int(rc.modelling_context.test_size * len(dataset))

    dataset = dataset.shuffle(buffer_size=10, reshuffle_each_iteration=False)

    train_dataset = dataset.take(train_size)
    val_dataset = dataset.skip(train_size)
    val_dataset = dataset.take(val_size)

    per_device_batch_size = rc.modelling_context.per_device_batch_size

    per_host_batch_size = per_device_batch_size * jax.local_device_count()

    train_dataset = train_dataset.batch(per_host_batch_size, drop_remainder=False)
    test_dataset = test_dataset.batch(per_host_batch_size, drop_remainder=False)

    hps = HyperParameters()
    hps.number_epochs = rc.modelling_context.number_epochs
    hps.num_classes = rc.modelling_context.number_classes
    hps.patch_size = rc.modelling_context.patch_size
    hps.num_channels = rc.modelling_context.number_channels

    model, results = train_and_evaluate(
        hyper_parameters=hps,
        runtime_context=rc,
        train_dataset=train_dataset,
        val_dataset=val_dataset,
    )
    return rc



from pathlib import Path


from dataclasses import dataclass
from ...d_utils.enums import Step


@dataclass
class DataSet:
    name: str
    is_in_separate_folder: bool

    def where_is_it(self, step: Step = None):
        return



################################################################################
# File: /best_package/d03_outputting/services/finalizing_srv.py                #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 5:45:31 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 5:45:31 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils.runtime_context import RuntimeContext
from ...d_utils import logger


def run(rc: RuntimeContext):
    logger.info("finalizing outputting step")
    return rc



################################################################################
# File: /best_package/d_utils/package_initializer.py                           #
# Project: best_project                                                        #
# File Created: Friday, 9th September 2022 3:36:40 am                          #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Friday, 9th September 2022 3:36:41 am                         #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from pathlib import Path
import os
from ..d_utils.enums import *
import yaml
import sys
from datetime import datetime
from . import tools as tl


def config_generator(conf_type: Config_type):
    conf = {}
    if conf_type == Config_type.base:
        conf = {
            "control": {
                "method": "method01",
                "data_access": {"data_set": "dataset01", "data_set_format": "images"},
                "pre_processed": {"save_data": True, "enabled": False},
                "modelled": {
                    "clear_last_run_data": False,
                    "save_data": True,
                    "enabled": False,
                },
                "outputted": {
                    "clear_last_run_data": False,
                    "save_data": True,
                    "enabled": False,
                },
            }
        }
    else:
        conf = {
            "root_data_path": "/volume/data/",
            "mrc_data_dir": "/volume/data/",
            "spartan_data_dir": "/data/projects/punim1484/data/",
        }

    return conf


def initialize_config(execuation_path: Path):
    config_folder_path = Path(execuation_path / "conf")
    base_config_path = Path(config_folder_path / "base.yaml")
    local_config_path = Path(config_folder_path / "local.yaml")
    try:
        if not os.path.exists(base_config_path):
            Path(config_folder_path).mkdir(parents=True, exist_ok=True)
            print("generating new base config file at %s" % base_config_path)
            with open(base_config_path, "w") as base_config_file:
                yaml.dump(
                    config_generator(Config_type.base),
                    base_config_file,
                    default_flow_style=False,
                    sort_keys=False,
                )

        if not os.path.exists(local_config_path):
            Path(config_folder_path).mkdir(parents=True, exist_ok=True)
            print("generating new local file at %s" % local_config_path)
            with open(local_config_path, "w") as local_config_file:
                yaml.dump(
                    config_generator(Config_type.local),
                    local_config_file,
                    default_flow_style=False,
                    sort_keys=False,
                )
    except Exception as e:
        print(e.__str__())
    return True


def generate_first_notebook(notebook_folder_path: Path):
    import getpass

    import nbformat as nbf

    nb = nbf.v4.new_notebook()
    cells = []

    # Assume make_image() saves an image to file and returns the filename
    # image_file = make_image(var)
    text = "Hello"

    cell = nbf.v4.new_code_cell(
        source=[
            "import os",
            "import sys",
            "module_path = os.path.abspath(os.path.join('..'))",
            "sys.path.append(module_path)",
        ]
    )

    cell2 = nbf.v4.new_code_cell(
        source=["import best_package", "best_package.initialise()"]
    )

    cells.append(cell)
    cells.append(cell2)

    nb["cells"] = cells

    notebook_path = str(notebook_folder_path / "%s_%s_EDA.ipynb") % (
        datetime.now().strftime("%Y_%m_%d"),
        getpass.getuser(),
    )

    with open(str(notebook_path), "a+") as f:
        nbf.write(nb, f)

    return


def initialize_notebooks(execuation_path: Path):

    notebooks_path = execuation_path / "notebooks"

    if not os.path.exists(notebooks_path):
        print("making the notebooks folder at %s" % notebooks_path)
        Path(notebooks_path).mkdir(parents=True, exist_ok=True)
        generate_first_notebook(notebooks_path)
    else:
        notebook_folder_empty = True
        with os.scandir(notebooks_path) as it:
            if any(it):
                notebook_folder_empty = False
        if notebook_folder_empty:
            generate_first_notebook(notebooks_path)

    return


def run(execuation_path: Path):
    is_running_in_notebook = tl.isnotebook()
    if is_running_in_notebook:
        print("running in notebook")
        execuation_path = Path(os.getcwd()).parents[0]
    print("running from %s" % execuation_path)
    initialize_config(execuation_path=execuation_path)
    initialize_notebooks(execuation_path=execuation_path)
    return



################################################################################
# File: /best_package/d00_data_access/model/contextmodel.py                    #
# Project: best_project                                                        #
# File Created: Friday, 9th September 2022 1:18:41 am                          #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Friday, 9th September 2022 1:18:41 am                         #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################

from .datamodel import *
from ...d_utils.enums import *
from typing import List
from ...d_utils.data_browser import DataBrowser
import os
import pathlib as pl


class DataAccessContext:
    data_sets: List[DataSet]
    data_browser: DataBrowser
    development_state: Development_state

    def __init__(self):
        pass

    def check_input_datasets_exist(self):
        for data_set in self.data_sets:
            data_path = self.data_browser.top_original_path(data_set=data_set)
            # check if there is data there
            data_size = 0
            for dirpath, dirnames, filenames in os.walk(data_path):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    if not os.path.islink(fp):
                        data_size += os.path.getsize(fp)
            if data_size == 0:
                raise RuntimeError(
                    "data set %s is empty. Make sure that you have the original data for %s placed at %s"
                    % (data_set.name, data_set.name, data_path)
                )
        return True

    def make_working_folder(self):
        try:
            pl.Path(self.data_browser.top_working_path(mode=Mode.development)).mkdir(
                parents=True, exist_ok=True
            )
            pl.Path(self.data_browser.top_working_path(mode=Mode.production)).mkdir(
                parents=True, exist_ok=True
            )
        except Exception as e:
            raise e
        return 0

    def make_output_folder(self):
        try:

            pl.Path(self.data_browser.top_output_path(mode=Mode.development)).mkdir(
                parents=True, exist_ok=True
            )
            pl.Path(self.data_browser.top_output_path(mode=Mode.production)).mkdir(
                parents=True, exist_ok=True
            )
        except Exception as e:
            raise e
        return 0

    def run(self):
        if self.development_state != Development_state.skeleton:
            self.check_input_datasets_exist()
        return 0

    def make_it(self):
        self.make_output_folder()
        self.make_working_folder()
        return 0



################################################################################
# File: /best_package/d_utils/file_names.py                                    #
# Project: best_project                                                        #
# File Created: Wednesday, 14th September 2022 6:52:10 am                      #
# Author: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                          #
# ---------------------------------------------------------------              #
# Last Modified: Wednesday, 14th September 2022 6:52:10 am                     #
# Modified By: Zaher Joukhadar (zjoukhadar@unimelb.edu.au)                     #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from .constants import *


def modelling_context_file_name(run_name: str):
    return "modelling_context_" + run_name + CONTEXT_FILE_EXTENTION


def pre_processing_context_file_name():
    return "pre_processing_context" + CONTEXT_FILE_EXTENTION



################################################################################
# File: /tools.py                                                              #
# Project: d_utils                                                             #
# File Created: Sunday, 10th October 2021 4:10:22 am                           #
# Author: Ubuntu# zjoukhadar@unimelb.edu.au                                    #
# ---------------------------------------------------------------              #
# Last Modified: Sunday, 10th October 2021 4:10:22 am                          #
# Modified By: Ubuntu# zjoukhadar@unimelb.edu.au                               #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
################################################################################
# File: /tools.py                                                              #
# Project: d_utils                                                             #
# File Created: Monday, 12th April 2021 4:13:37 am                             #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Monday, 12th April 2021 4:13:38 am                            #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
import multiprocessing, logging
from pathlib import Path
import os
from IPython import get_ipython
from ..d_utils.enums import *
import getpass
import yaml

from . import logger


def root():
    return str(Path(__file__).parent.parent.parent)


def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def save_context(context, step: Step):

    return


def get_context(step: Step, run_time: str = None):
    return


# region data_store tool functions:


# endregion


# endregion



################################################################################
# File: /data_browser.py                                                       #
# Project: d_utils                                                             #
# File Created: Sunday, 10th October 2021 9:10:48 am                           #
# Author: Ubuntu# zjoukhadar@unimelb.edu.au                                    #
# ---------------------------------------------------------------              #
# Last Modified: Sunday, 10th October 2021 9:10:49 am                          #
# Modified By: Ubuntu# zjoukhadar@unimelb.edu.au                               #
# ---------------------------------------------------------------              #
# Copyright 2021 - 2021, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from .enums import Step, Method, Mode
import getpass
from .. import get_version, get_package_name
from ..d00_data_access.model.datamodel import DataSet
from pathlib import Path


class DataBrowser:
    original_data_path: str
    root_data_path: str

    method: Method

    def __init__(
        self,
        original_data_path: str,
        root_data_path: str,
    ):
        self.root_data_path = root_data_path
        self.original_data_path = original_data_path

    def top_original_path(self, data_set: DataSet):
        if data_set.is_in_separate_folder:
            return self.original_data_path + data_set.name
        else:
            return self.original_data_path

    def top_working_path(self, mode: Mode):
        top_user_working_path = (
            self.root_data_path
            + ((getpass.getuser() + "/") if mode == Mode.development else "")
            + get_package_name()
            + "/"
            + str(get_version())
            + "/working/"
        )
        return top_user_working_path

    def step_working_path(self, step: Step, mode: Mode, data_set: DataSet = None):
        step_user_working_path = (
            self.top_working_path(mode=mode)
            + step.name
            + "/"
            + (data_set.name + "/" if data_set else "")
        )
        return step_user_working_path

    def run_modelling_path(
        self,
        mode: Mode,
        run_name: str,
        data_set: DataSet = None,
    ):
        run_user_working_path = (
            self.step_working_path(Step.modelled, mode, data_set) + run_name + "/"
        )
        return run_user_working_path

    def top_output_path(self, mode: Mode):
        top_output_path = (
            self.root_data_path
            + ((getpass.getuser() + "/") if mode == Mode.development else "")
            + "/"
            + get_package_name()
            + "/"
            + str(get_version())
            + "/output/"
        )
        return top_output_path

    def run_output_path(self, mode: Mode, run_name: str, data_set: DataSet = None):
        run_output_path = (
            self.top_output_path(mode=mode)
            + run_name
            + "/"
            + (data_set.name + "/" if data_set else "")
        )
        return run_output_path



################################################################################
# File: /deepeye/d06_reporting/services/initialise_srv.py                      #
# Project: deep_eye_no_dementia                                                #
# File Created: Friday, 20th May 2022 6:00:08 am                               #
# Author: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                           #
# ---------------------------------------------------------------              #
# Last Modified: Friday, 20th May 2022 6:00:08 am                              #
# Modified By: Zaher Joukhadar# zjoukhadar@unimelb.edu.au                      #
# ---------------------------------------------------------------              #
# Copyright 2022 - 2022, Melbourne Data Analytics Platform (MDAP)              #
################################################################################
from ...d_utils.runtime_context import RuntimeContext
from ...d_utils import logger
from ...d_utils.enums import *
from ...d_utils.cleaners import clear_last_run_data
import pickle
from ...d_utils import file_names


def initialise_outputting(rc: RuntimeContext):
    from ..model.contextmodel import OutputtingContext

    logger.info("initialising outputting step")
    last_run_name = rc.get_previous_run_name(Step.outputted)
    modeller_run_name = rc.config.what_run_name_to_use_in_output()

    rc.set_current_run_name(modeller_run_name, Step.outputted)

    my_outputting_context = OutputtingContext(
        modeller_run_name=modeller_run_name, last_run_name=last_run_name
    )

    my_outputting_context.data_browser = rc.data_browser
    my_outputting_context.mode = rc.config.get_mode(Step.outputted)
    my_outputting_context.development_state = rc.config.get_development_state()

    data_access_data_sets = rc.config.get_data_sets()
    outputting_data_sets = rc.config.get_data_sets(step=Step.outputted)

    my_outputting_context.data_sets = [
        da_ds
        for pp_ds in outputting_data_sets
        for da_ds in data_access_data_sets
        if da_ds.name == pp_ds
    ]

    logger.info("output using run name: %s" % (modeller_run_name))

    try:
        my_outputting_context.run()
        my_outputting_context.make_it()
    except Exception as e:
        raise e
    rc.outputting_context = my_outputting_context
    if (
        rc.config.should_clear_last_output()
        and last_run_name != rc.config.keep_this_output_data()
    ):
        clear_last_run_data(rc, Step.outputted)

    # retrieve modelling context
    context_file_path = rc.modelling_context.get_modelling_folder_path(
        modeller_run_name
    ) + file_names.modelling_context_file_name(modeller_run_name)

    with open(context_file_path, "rb") as handle:
        rc.modelling_context = pickle.load(handle)
    return rc



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['best_package',
 'best_package.d00_data_access',
 'best_package.d00_data_access.datasets.dataset01_dataset',
 'best_package.d00_data_access.model',
 'best_package.d00_data_access.services',
 'best_package.d01_pre_processing',
 'best_package.d01_pre_processing.model',
 'best_package.d01_pre_processing.services',
 'best_package.d02_modelling',
 'best_package.d02_modelling.model',
 'best_package.d02_modelling.services',
 'best_package.d03_outputting',
 'best_package.d03_outputting.model',
 'best_package.d03_outputting.services',
 'best_package.d_utils']

package_data = \
{'': ['*'],
 'best_package.d00_data_access.datasets.dataset01_dataset': ['dummy_data/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'colored>=1.4.3,<2.0.0',
 'flax>=0.6.0,<0.7.0',
 'ipython>=8.5.0,<9.0.0',
 'ml-collections>=0.1.1,<0.2.0',
 'nbformat>=5.4.0,<6.0.0',
 'notebook>=6.4.12,<7.0.0',
 'optax>=0.1.3,<0.2.0',
 'pandas>=1.4.4,<2.0.0',
 'pyfiglet>=0.8.post1,<0.9',
 'tensorflow-datasets>=4.6.0,<5.0.0',
 'tensorflow>=2.10.0,<3.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'best-package',
    'version': '0.1.1',
    'description': '',
    'long_description': '',
    'author': 'Zaher Joukhadar',
    'author_email': 'zaher@joukhadar.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import imp
from ...d_utils.runtime_context import RuntimeContext
from ...d_utils.cleaners import clear_last_run_data

from ...d_utils import logger
from ...d_utils.enums import *
import os


def initialise_modeller(rc: RuntimeContext):
    logger.info("initialising modelling step")

    from ..model.contextmodel import ModellingContext

    last_run_name = rc.get_previous_run_name(Step.modelled)
    if last_run_name is None or last_run_name == 0:
        current_run_name = "run_01"
    else:
        current_run_name = "run_" + str(int(last_run_name[4:]) + 1)
    rc.set_current_run_name(current_run_name, Step.modelled)
    my_modelling_context = ModellingContext(current_run_name, last_run_name)
    my_modelling_context.mode = rc.config.get_mode(Step.modelled)
    my_modelling_context.data_browser = rc.data_browser
    my_modelling_context.development_state = rc.config.get_development_state()
    my_modelling_context.test_size = rc.config.get_test_size()
    my_modelling_context.problem_type = rc.config.get_problem_type()
    my_modelling_context.number_classes = rc.config.get_number_of_classes()
    my_modelling_context.number_epochs = rc.config.get_number_epochs()
    my_modelling_context.per_device_batch_size = rc.config.get_per_device_batch_size()
    my_modelling_context.patch_size = rc.config.get_patch_size()
    my_modelling_context.number_channels = rc.config.get_number_of_channels()
    data_access_data_sets = rc.config.get_data_sets()
    modelling_data_sets = rc.config.get_data_sets(step=Step.modelled)

    my_modelling_context.data_sets = [
        da_ds
        for pp_ds in modelling_data_sets
        for da_ds in data_access_data_sets
        if da_ds.name == pp_ds
    ]

    my_modelling_context.method = rc.config.get_method(Step.modelled)

    logger.info(
        "current modelling run name: %s using %s"
        % (current_run_name, my_modelling_context.method.name)
    )

    try:
        my_modelling_context.check_it()
        my_modelling_context.make_it()
    except Exception as e:
        raise e
    rc.modelling_context = my_modelling_context

    if (
        rc.config.should_clear_last_run()
        and last_run_name != rc.config.keep_this_run_data()
    ):
        clear_last_run_data(rc, Step.modelled)

    os.environ["TFDS_DATA_DIR"] = rc.pre_processing_context.get_datasets_path()

    return rc
